import { GraphType, LinkType, NodeType } from '../types';
import { dataColors, visualizationColors } from 'config';
import { ReactEventHandler, useEffect, useImperativeHandle, useMemo, useRef, useState } from 'react';
import { Application, Circle, Color, Container, FederatedPointerEvent, Graphics, IPointData } from "pixi.js";
import { useAppDispatch, useML, useSearchResultData } from '../../../../data-access';
import { NLPopup } from './NLPopup';
import { hslStringToHex, nodeColor } from './utils';
import { CytoscapeLayout, GraphologyLayout, LayoutFactory, Layouts } from '../../../../graph-layout';
import { MultiGraph } from 'graphology';
import { VisualizationConfiguration } from '../../../common';
import { Viewport } from 'pixi-viewport';

type Props = {
  onClick: (node: NodeType, pos: IPointData) => void;
  // onHover: (data: { node: NodeType; pos: IPointData }) => void;
  // onUnHover: (data: { node: NodeType; pos: IPointData }) => void;
  highlightNodes: NodeType[];
  configuration: VisualizationConfiguration;
  currentShortestPathEdges?: LinkType[];
  highlightedLinks?: LinkType[];
  graph?: GraphType;
  layoutAlgorithm: string;
};

const app = new Application({
  background: 0xffffff,
  antialias: true,
  autoDensity: true,
  eventMode: 'auto',
  resolution: window.devicePixelRatio || 1,
});
const nodeLayer = new Container();
const linkLayer = new Container();

type LayoutState = 'reset' | 'running' | 'paused';

//////////////////
// MAIN COMPONENT
//////////////////

export const NLPixi = (props: Props) => {
  const [quickPopup, setQuickPopup] = useState<{ node: NodeType; pos: IPointData } | undefined>();
  const [popups, setPopups] = useState<{ node: NodeType; pos: IPointData }[]>([]);

  const nodeMap = useRef(new Map<string, Graphics>());
  const linkMap = useRef(new Map<string, Graphics>());
  const viewport = useRef<Viewport>();
  const layoutState = useRef<LayoutState>('reset');
  const layoutStoppedCount = useRef(0);
  const ref = useRef<HTMLDivElement>(null);
  const mouseInCanvas = useRef<boolean>(false);
  const isSetup = useRef(false);
  const ml = useML();
  const dragging = useRef<{ node: NodeType; gfx: Graphics } | null>(null);
  const onlyClicked = useRef(false);
  const dispatch = useAppDispatch();
  const searchResults = useSearchResultData();

  const layoutAlgorithm = useRef<CytoscapeLayout | GraphologyLayout>(new LayoutFactory().createLayout(Layouts.DAGRE));

  // const cull = new Cull();
  // let cullDirty = useRef(true);

  const [config, setConfig] = useState({
    width: 1000,
    height: 1000,

    LAYOUT_ALGORITHM: Layouts.FORCEATLAS2WEBWORKER,

    NODE_RADIUS: 5,
    NODE_BORDER_LINE_WIDTH: 1.0,
    NODE_BORDER_LINE_WIDTH_SELECTED: 5.0, // if selected and normal width are different the thicker line will be still in the gfx
    NODE_BORDER_COLOR_DEFAULT: dataColors.neutral[70],
    NODE_BORDER_COLOR_SELECTED: dataColors.orange[60],

    LINE_COLOR_DEFAULT: dataColors.neutral[40],
    LINE_COLOR_SELECTED: visualizationColors.GPSelected.colors[1],
    LINE_COLOR_ML: dataColors.blue[60],
    LINE_WIDTH_DEFAULT: 0.8,
  });

  useEffect(() => {
    setConfig((lastConfig) => {
      return {
        ...lastConfig,
        LAYOUT_ALGORITHM: (props.layoutAlgorithm as Layouts) || lastConfig.LAYOUT_ALGORITHM,
      };
    });
  }, [props.layoutAlgorithm]);

  const imperative = useRef<any>(null);

  useImperativeHandle(imperative, () => ({
    onDragStart(node: NodeType, gfx: Graphics) {
      // todo: graphology does not support fixed nodes
      // todo: after vis-settings panel is there, we should to also support the original d3 force to allow interactivity if needed
      if (props.layoutAlgorithm === Layouts.FORCEATLAS2WEBWORKER) return;
      if (viewport.current) viewport.current.pause = true;
      dragging.current = { node, gfx };
      onlyClicked.current = true;
    },

    onDragMove(movementX: number, movementY: number) {
      if (dragging.current) {
        onlyClicked.current = false;
        if (quickPopup) setQuickPopup(undefined);
        const idx = popups.findIndex((p) => p.node.id === dragging.current?.node.id);
        if (idx >= 0) {
          const p = popups[idx];
          p.pos.x += movementX / (viewport.current?.scaled || 1);
          p.pos.y += movementY / (viewport.current?.scaled || 1);
          popups[idx] = p;
          setPopups([...popups]);
        }

        if (!dragging.current.node.fx) dragging.current.node.fx = dragging.current.node.x || 0;
        if (!dragging.current.node.fy) dragging.current.node.fy = dragging.current.node.y || 0;
        dragging.current.node.fx += movementX / (viewport.current?.scaled || 1);
        dragging.current.node.fy += movementY / (viewport.current?.scaled || 1);
        // force.simulation.alpha(0.1).restart();
      }
    },

    onDragEnd() {
      if (dragging.current) {
        // dragging.current.node.fx = null;
        // dragging.current.node.fy = null;
        if (viewport.current) viewport.current.pause = false;
        if (onlyClicked.current) {
          onlyClicked.current = false;

          if (popups.filter((d) => d.node.id === dragging.current?.node.id).length > 0) {
            setPopups(popups.filter((p) => p.node.id !== dragging.current?.node.id));
          } else {
            console.log('clicked', popups);
            setPopups([...popups, { node: dragging.current.node, pos: toGlobal(dragging.current.node) }]);
          }

          props.onClick(dragging.current.node, toGlobal(dragging.current.node));
        }
        this.onHover(dragging.current.node);
        dragging.current = null;
      }
    },
    onHover(node: NodeType) {
      if (
        mouseInCanvas.current &&
        viewport?.current &&
        !viewport?.current?.pause &&
        node &&
        popups.filter((p) => p.node.id === node.id).length === 0
      ) {
        setQuickPopup({ node, pos: toGlobal(node) });
      }
    },
    onUnHover() {
      setQuickPopup(undefined);
    },
    onPan() {
      setPopups([]);
    },
  }));

  function resize() {
    const width = ref?.current?.clientWidth || 1000;
    const height = ref?.current?.clientHeight || 1000;
    app.renderer.resize(width, height);
    if (viewport.current) {
      viewport.current.screenWidth = width;
      viewport.current.worldWidth = width;
      viewport.current.worldHeight = height;
      viewport.current.screenHeight = height;
    }
    app.render();
  }

  useEffect(() => {
    if (!ref.current) return;
    const resizeObserver = new ResizeObserver(() => {
      resize();
    });
    resizeObserver.observe(ref.current);
    return () => resizeObserver.disconnect(); // clean up
  }, []);

  useEffect(() => {
    if (ref.current && ref.current.children.length === 0) {
      ref.current.appendChild(app.view as HTMLCanvasElement);
      resize();
    }
  }, [ref]);

  function toGlobal(node: NodeType): IPointData {
    if (viewport?.current) {
      // const rect = ref.current?.getBoundingClientRect();
      const rect = { x: 0, y: 0 };
      const x = (rect?.x || 0) + (node.x || 0);
      const y = (rect?.y || 0) + (node.y || 0);
      return viewport.current.toScreen(x, y);
    } else return { x: 0, y: 0 };
  }

  function onDragStart(event: FederatedPointerEvent, node: NodeType, gfx: Graphics) {
    event.stopPropagation();
    if (imperative.current) imperative.current.onDragStart(node, gfx);
  }

  function onDragMove(event: FederatedPointerEvent) {
    event.stopPropagation();
    if (imperative.current) imperative.current.onDragMove(event.movementX, event.movementY);
  }

  function onDragEnd(event: FederatedPointerEvent) {
    event.stopPropagation();
    if (imperative.current) imperative.current.onDragEnd();
  }

  const updateNode = (node: NodeType) => {
    const gfx = nodeMap.current.get(node.id);
    if (!gfx) return;

    const lineColor = node.isShortestPathSource
      ? dataColors.orange['95']
      : node.isShortestPathTarget
        ? dataColors.blue['95']
        : node.selected
          ? config.NODE_BORDER_COLOR_SELECTED
          : config.NODE_BORDER_COLOR_DEFAULT;
    const lineWidth = node.selected ? config.NODE_BORDER_LINE_WIDTH_SELECTED : config.NODE_BORDER_LINE_WIDTH;
    gfx.lineStyle(lineWidth, new Color(hslStringToHex(lineColor)));

    if (node?.cluster) {
      gfx.beginFill(node.cluster >= 0 ? nodeColor(node.cluster) : 0x000000);
    } else gfx.beginFill(nodeColor(node.type));

    gfx.drawCircle(0, 0, Math.max(node.radius || 5, 5));
    gfx.endFill();

    gfx.position.set(node.x, node.y);

    gfx.off('mouseover');
    gfx.off('mousedown');
    gfx.on('mouseover', (e) => {
      e.stopPropagation();
      e.preventDefault();
      if (imperative.current) imperative.current.onHover(node);
    });
    gfx.on('mouseout', (e) => {
      e.stopPropagation();
      e.preventDefault();
      if (imperative.current) imperative.current.onUnHover();
    });
    gfx.on('mousedown', (e) => onDragStart(e, node, gfx));

    // if (!item.position) {
    //   item.position = new Point(node.x, node.y);
    // } else {
    //   item.position.set(node.x, node.y);
    // }
    // Update attributes position if they exist
    // if (node.gfxAttributes) {
    //   const x = node.x - node.gfxAttributes.width / 2;
    //   const y = node.y - node.gfxAttributes.height - 20;
    //   if (!node.gfxAttributes?.position) node.gfxAttributes.position = new Point(x, y);
    //   else {
    //     node.gfxAttributes.position.set(x, y);
    //   }
    // }
  };

  const createNode = (node: NodeType, selected?: boolean) => {
    // check if node is already drawn, and if so, delete it
    if (node && node?.id && nodeMap.current?.has(node.id)) {
      nodeMap.current.delete(node.id);
    }
    // Do not draw node if it has no position
    if (node.x === undefined || node.y === undefined) return;
    const gfx = new Graphics();
    nodeMap.current.set(node.id, gfx);
    nodeLayer.addChild(gfx);
    node.selected = selected;

    updateNode(node);
    gfx.name = 'node_' + node.id;
    gfx.eventMode = 'dynamic';

    return gfx;
  };

  // /** UpdateRadius works just like UpdateColors, but also applies radius*/
  // const UpdateRadius = (graph: GraphType, radius: number) => {
  //   // update for each node in graph
  //   graph.nodes.forEach((node: NodeType) => {
  //     createNode(node);
  //   });
  // };

  const updateLink = (link: LinkType, gfx: Graphics) => {
    if (!props.graph) return;

    const _source = link.source;
    const _target = link.target;

    if (!_source || !_target) {
      console.log('source or target not found', _source, _target);
      return;
    }

    let sourceId = '';
    let targetId = '';
    let source: NodeType | undefined;
    let target: NodeType | undefined;

    if (typeof _source === 'string') {
      sourceId = link.source as string;
      targetId = link.target as string;
      source = nodeMap.current.get(sourceId) as NodeType | undefined;
      target = nodeMap.current.get(targetId) as NodeType | undefined;
    } else {
      source = link.source as NodeType;
      target = link.target as NodeType;
      sourceId = source.id;
      targetId = target.id;
    }
    if (!source || !target) {
      console.error('source or target not found', source, target, sourceId, targetId);
      return;
    }

    if (gfx) {
      // let color = link.color || 0x000000;
      let color = config.LINE_COLOR_DEFAULT;
      let style = config.LINE_WIDTH_DEFAULT;
      let alpha = link.alpha || 1;
      if (link.mlEdge) {
        color = config.LINE_COLOR_ML;
        if (link.value > ml.communityDetection.jaccard_threshold) {
          style = link.value * 1.8;
        } else {
          style = 0;
          alpha = 0.2;
        }
      } else if (props.highlightedLinks && props.highlightedLinks.includes(link)) {
        if (link.mlEdge && ml.communityDetection.jaccard_threshold) {
          if (link.value > ml.communityDetection.jaccard_threshold) {
            color = dataColors.magenta[50];
            // 0xaa00ff;
            style = link.value * 1.8;
          }
        } else {
          color = dataColors.red[70];
          // color = 0xff0000;
          style = 1.0;
        }
      } else if (props.currentShortestPathEdges && props.currentShortestPathEdges.includes(link)) {
        color = dataColors.green[50];
        // color = 0x00ff00;
        style = 3.0;
      }

      gfx.clear();
      gfx.beginFill();
      gfx
        .lineStyle(style, hslStringToHex(color), alpha)
        .moveTo(source.x || 0, source.y || 0)
        .lineTo(target.x || 0, target.y || 0);
      gfx.endFill();
    } else {
      throw Error('Link not found');
    }
  };

  const createLink = (link: LinkType) => {
    if (linkMap.current.has(link.id)) {
      linkMap.current.delete(link.id);
    }
    const gfx = new Graphics();
    gfx.name = 'link_' + link.id;
    linkMap.current.set(link.id, gfx);
    updateLink(link, gfx);
    linkLayer.addChild(gfx);
    return gfx;
  };

  useEffect(() => {
    return () => {
      nodeMap.current.clear();
      linkMap.current.clear();
      nodeLayer.removeChildren();
      linkLayer.removeChildren();
    };
  }, []);

  useEffect(() => {
    if (props.graph && ref.current && ref.current.children.length > 0 && imperative.current) {
      if (isSetup.current === false) setup();
      else update(false);
    }
  }, [props.graph, config]);

  useEffect(() => {
    if (props.graph) {
      props.graph.nodes.forEach((node: NodeType) => {
        const gfx = nodeMap.current.get(node.id);
        if (!gfx) return;
        const isNodeInSearchResults = searchResults.nodes.some((resultNode) => resultNode.id === node.id);
        gfx.alpha = isNodeInSearchResults || searchResults.nodes.length === 0 ? 1 : 0.05;
      });

      props.graph.links.forEach((link: LinkType) => {
        const gfx = linkMap.current.get(link.id);
        if (!gfx) return;
        const isLinkInSearchResults = searchResults.edges.some((resultEdge) => resultEdge.id === link.id);
        gfx.alpha = isLinkInSearchResults || searchResults.edges.length === 0 ? 1 : 0.05;
      });
    }
  }, [searchResults]);

  const tick = (delta: number) => {
    if (layoutState.current === 'paused') return;
    if (layoutState.current === 'reset') layoutStoppedCount.current = 0;

    if (props.graph) {
      if (!layoutAlgorithm.current) return;

      let stopped = 0;
      props.graph.nodes.forEach((node: NodeType, i) => {
        if (!layoutAlgorithm.current) return;
        const gfx = nodeMap.current.get(node.id);
        if (!gfx || node.x === undefined || node.y === undefined) return;

        const position = layoutAlgorithm.current.getNodePosition(node.id);

        if (Math.abs(node.x - position.x - app.renderer.width / 2) + Math.abs(node.y - position.y - app.renderer.height / 2) < 1) {
          stopped += 1;
          return;
        }
        if (layoutAlgorithm.current.provider === 'Graphology') {
          // this is a dirty hack to fix the graphology layout being out of bounds
          node.x = position.x + app.renderer.width / 2;
          node.y = position.y + app.renderer.height / 2;
        } else {
          node.x = position.x;
          node.y = position.y;
        }

        if (layoutState.current === 'running') {
          gfx.position.copyFrom({
            x: node.fx || gfx.position.x + ((node.x || 0) - gfx.position.x) * 0.1 * delta,
            y: node.fy || gfx.position.y + ((node.y || 0) - gfx.position.y) * 0.1 * delta,
          });
        } else {
          gfx.position.copyFrom(node as IPointData);
        }
      });

      if (stopped === props.graph.nodes.length) {
        layoutStoppedCount.current = layoutStoppedCount.current + 1;
        if (layoutStoppedCount.current > 1000) {
          layoutState.current = 'paused';
          console.debug('NL layout paused');
        }
      } else {
        layoutStoppedCount.current = 0;
      }
      if (layoutState.current === 'reset') {
        layoutState.current = 'running';
      }

      // Update forces of the links
      props.graph.links.forEach((link: any) => {
        if (linkMap.current && !!linkMap.current.has(link.id)) {
          updateLink(link, linkMap.current.get(link.id) as Graphics);
        }
      });
    }
  };

  const update = (forceClear = false) => {
    setPopups([]);
    if (!props.graph || !ref.current) return;

    if (props.graph) {
      if (forceClear) {
        nodeMap.current.clear();
        linkMap.current.clear();
        nodeLayer.removeChildren();
        linkLayer.removeChildren();
      }

      nodeMap.current.forEach((gfx, id) => {
        if (!props.graph?.nodes?.find((node) => node.id === id)) {
          nodeLayer.removeChild(gfx);
          gfx.destroy();
          nodeMap.current.delete(id);
        }
      });

      linkMap.current.forEach((gfx, id) => {
        if (!props.graph?.links?.find((link) => link.id === id)) {
          linkLayer.removeChild(gfx);
          gfx.destroy();
          linkMap.current.delete(id);
        }
      });

      props.graph.nodes.forEach((node: NodeType) => {
        if (!forceClear && nodeMap.current.has(node.id)) {
          const old = nodeMap.current.get(node.id);
          node.x = old?.x || node.x;
          node.y = old?.y || node.y;
          updateNode(node);
        } else {
          createNode(node);
        }
      });
      props.graph.links.forEach((link: LinkType) => {
        if (!forceClear && linkMap.current.has(link.id)) {
          const gfx = linkMap.current.get(link.id);
          if (gfx) updateLink(link, gfx);
        } else createLink(link);
      });

      // // update text colour (written after nodes so that text appears on top of nodes)
      //   nodes.forEach((node: NodeType) => {
      //   if (node.gfxAttributes !== undefined) {
      //       const selected = node.selected === true;
      //       node.gfxAttributes.destroy();
      //       createAttributes(node);
      //       if (selected) {
      //       showAttributes(node);
      //       }
      //   }
      //   });
      // console.log(nodes.children);

      // force.startSimulation(props.graph, ref.current.getBoundingClientRect());
      // force.simulation.on('tick', () => {});
      app.ticker.add(tick);
      layoutState.current = 'reset';
      if (forceClear) setupLayout(forceClear);
    }
  };

  /**
   * SetNodeGraphics is an initializing function. It attaches the nodes and links to the simulation.
   * It creates graphic objects and adds these to the PIXI containers. It also clears both of these of previous nodes and links.
   * @param graph The graph returned from the database and that is parsed into a nodelist and edgelist.
   */
  const setup = () => {
    nodeLayer.removeChildren();
    linkLayer.removeChildren();
    app.stage.removeChildren();

    if (!props.graph) throw Error('Graph is undefined');

    const size = ref.current?.getBoundingClientRect();
    viewport.current = new Viewport({
      screenWidth: size?.width || 1000,
      screenHeight: size?.height || 1000,
      worldWidth: size?.width || 1000,
      worldHeight: size?.height || 1000,
      stopPropagation: true,
      events: app.renderer.events, // the interaction module is important for wheel to work properly when renderer.view is placed or scaled
    });
    app.stage.addChild(viewport.current);
    // activate plugins
    viewport.current.drag().pinch().wheel({ smooth: 2 }).animate({}).decelerate({ friction: 0.75 });

    viewport.current.addChild(linkLayer);
    viewport.current.addChild(nodeLayer);
    viewport.current.on('drag-start', (event) => {
      imperative.current.onPan();
    });

    app.stage.eventMode = 'dynamic';
    app.stage.on('mouseup', onDragEnd);
    app.stage.on('pointerup', onDragEnd);
    app.stage.on('mousemove', onDragMove);
    app.stage.on('mouseup', onDragEnd);

    nodeMap.current.clear();
    linkMap.current.clear();
    update(true);
    isSetup.current = true;
  };

  const setupLayout = (forceClear: boolean) => {
    const layoutFactory = new LayoutFactory();
    layoutAlgorithm.current = layoutFactory.createLayout(config.LAYOUT_ALGORITHM);

    if (!layoutAlgorithm) throw Error('LayoutAlgorithm is undefined');

    const graphologyGraph = new MultiGraph();
    props.graph?.nodes.forEach((node) => {
      if (forceClear) graphologyGraph.addNode(node.id, { size: node.radius || 5 });
      else graphologyGraph.addNode(node.id, { size: node.radius || 5, x: node.x || 0, y: node.y || 0 });
    });

    props.graph?.links.forEach((link) => {
      graphologyGraph.addEdge(link.source, link.target);
    });
    const boundingBox = { x1: 0, x2: app.renderer.screen.width, y1: 0, y2: app.renderer.screen.height };

    if (forceClear) {
      const startingLayout = layoutFactory.createLayout('Graphology_random');
      startingLayout.layout(graphologyGraph, boundingBox);
    }
    layoutAlgorithm.current.layout(graphologyGraph, boundingBox);
  };
  return (
    <>
      {mouseInCanvas.current && popups.map((popup) => <NLPopup onClose={() => {}} data={popup} key={popup.node.id} />)}
      {quickPopup && <NLPopup onClose={() => {}} data={quickPopup} />}
      <div
        className="h-full w-full overflow-hidden"
        ref={ref}
        onMouseEnter={(e) => {
          mouseInCanvas.current = true;
        }}
        onMouseOut={(e) => {
          mouseInCanvas.current = false;
        }}
      ></div>
    </>
  );
};
